@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2012 inria.  All rights reserved.
+ * Copyright © 2009-2014 Inria.  All rights reserved.
  * Copyright © 2009-2011 Université Bordeaux 1
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -27,43 +27,32 @@ hwloc__xml_verbose(void)
   return verbose;
 }
 
-/***********************************
- ******** Backend Init/Exit ********
- ***********************************/
+/*********************************
+ ********* XML callbacks *********
+ *********************************/
 
-/* this can be the first XML call */
-int
-hwloc_backend_xml_init(struct hwloc_topology *topology, const char *xmlpath, const char *xmlbuffer, int xmlbuflen)
-{
-  int ret;
-#ifdef HWLOC_HAVE_LIBXML2
-  char *env = getenv("HWLOC_NO_LIBXML_IMPORT");
-  if (!env || !atoi(env)) {
-    ret = hwloc_libxml_backend_init(topology, xmlpath, xmlbuffer, xmlbuflen);
-  } else
-#endif /* HWLOC_HAVE_LIBXML2 */
-  {
-    ret = hwloc_nolibxml_backend_init(topology, xmlpath, xmlbuffer, xmlbuflen);
-  }
-  if (ret < 0)
-    return ret;
+/* set when registering nolibxml and libxml components.
+ * modifications protected by the components mutex.
+ * read by the common XML code in topology-xml.c to jump to the right XML backend.
+ */
+static struct hwloc_xml_callbacks *hwloc_nolibxml_callbacks = NULL, *hwloc_libxml_callbacks = NULL;
 
-  topology->is_thissystem = 0;
-  assert(topology->backend_type == HWLOC_BACKEND_NONE);
-  topology->backend_type = HWLOC_BACKEND_XML;
-
-  return 0;
-}
-
-/* this canNOT be the first XML call */
 void
-hwloc_backend_xml_exit(struct hwloc_topology *topology)
+hwloc_xml_callbacks_register(struct hwloc_xml_component *comp)
 {
-  assert(topology->backend_type == HWLOC_BACKEND_XML);
-  topology->backend_params.xml.backend_exit(topology);
-  topology->backend_type = HWLOC_BACKEND_NONE;
+  if (!hwloc_nolibxml_callbacks)
+    hwloc_nolibxml_callbacks = comp->nolibxml_callbacks;
+  if (!hwloc_libxml_callbacks)
+    hwloc_libxml_callbacks = comp->libxml_callbacks;
 }
 
+void
+hwloc_xml_callbacks_reset(void)
+{
+  hwloc_nolibxml_callbacks = NULL;
+  hwloc_libxml_callbacks = NULL;
+}			       
+
 /************************************************
  ********* XML import (common routines) *********
  ************************************************/
@@ -218,7 +207,7 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     switch (obj->type) {
     case HWLOC_OBJ_PCI_DEVICE:
     case HWLOC_OBJ_BRIDGE: {
-      obj->attr->pcidev.linkspeed = atof(value);
+      obj->attr->pcidev.linkspeed = (float) atof(value);
       break;
     }
     default:
@@ -424,7 +413,8 @@ hwloc__xml_import_pagetype(hwloc_topology_t topology __hwloc_attribute_unused, h
 }
 
 static int
-hwloc__xml_import_distances(hwloc_topology_t topology __hwloc_attribute_unused, hwloc_obj_t obj,
+hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
+			    hwloc_obj_t obj,
 			    hwloc__xml_import_state_t state)
 {
   unsigned long reldepth = 0, nbobjs = 0;
@@ -492,11 +482,11 @@ hwloc__xml_import_distances(hwloc_topology_t topology __hwloc_attribute_unused,
 
     distances->distances.latency_max = latmax;
 
-    if (topology->backend_params.xml.last_distances)
-      topology->backend_params.xml.last_distances->next = distances;
+    if (data->last_distances)
+      data->last_distances->next = distances;
     else
-      topology->backend_params.xml.first_distances = distances;
-    distances->prev = topology->backend_params.xml.last_distances;
+      data->first_distances = distances;
+    distances->prev = data->last_distances;
     distances->next = NULL;
   }
 
@@ -504,7 +494,63 @@ hwloc__xml_import_distances(hwloc_topology_t topology __hwloc_attribute_unused,
 }
 
 static int
-hwloc__xml_import_object(hwloc_topology_t topology, hwloc_obj_t obj,
+hwloc__xml_import_userdata(hwloc_topology_t topology __hwloc_attribute_unused, hwloc_obj_t obj,
+			   hwloc__xml_import_state_t state)
+{
+  size_t length = 0;
+  int encoded = 0;
+  char *name = NULL; /* optional */
+
+  while (1) {
+    char *attrname, *attrvalue;
+    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+      break;
+    if (!strcmp(attrname, "length"))
+      length = strtoul(attrvalue, NULL, 10);
+    else if (!strcmp(attrname, "encoding"))
+      encoded = !strcmp(attrvalue, "base64");
+    else if (!strcmp(attrname, "name"))
+      name = attrvalue;
+    else
+      return -1;
+  }
+
+  if (length && topology->userdata_import_cb) {
+    int ret;
+
+    if (encoded) {
+      char *encoded_buffer;
+      size_t encoded_length = 4*((length+2)/3);
+      ret = state->get_content(state, &encoded_buffer, encoded_length);
+      if (ret < 0)
+        return -1;
+      if (ret) {
+	char *decoded_buffer = malloc(length+1);
+	if (!decoded_buffer)
+	  return -1;
+	assert(encoded_buffer[encoded_length] == 0);
+	ret = hwloc_decode_from_base64(encoded_buffer, decoded_buffer, length+1);
+	if (ret != (int) length)
+	  return -1;
+	topology->userdata_import_cb(topology, obj, name, decoded_buffer, length);
+	free(decoded_buffer);
+      }
+    } else {
+      char *buffer;
+      ret = state->get_content(state, &buffer, length);
+      if (ret < 0)
+        return -1;
+      topology->userdata_import_cb(topology, obj, name, buffer, length);
+    }
+    state->close_content(state);
+  }
+  return state->close_tag(state);
+}
+
+static int
+hwloc__xml_import_object(hwloc_topology_t topology,
+			 struct hwloc_xml_backend_data_s *data,
+			 hwloc_obj_t obj,
 			 hwloc__xml_import_state_t state)
 {
   /* process attributes */
@@ -513,8 +559,7 @@ hwloc__xml_import_object(hwloc_topology_t topology, hwloc_obj_t obj,
     if (state->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "type")) {
-      obj->type = hwloc_obj_type_of_string(attrvalue);
-      if (obj->type == (hwloc_obj_type_t)-1)
+      if (hwloc_obj_type_sscanf(attrvalue, &obj->type, NULL, NULL, 0) < 0)
         return -1;
     } else {
       /* type needed first */
@@ -524,6 +569,12 @@ hwloc__xml_import_object(hwloc_topology_t topology, hwloc_obj_t obj,
     }
   }
 
+  if (obj->parent) {
+    /* root->parent is NULL, and root is already inserted */
+    hwloc_insert_object_by_parent(topology, obj->parent /* filled by the caller */, obj);
+    /* insert_object_by_parent() doesn't merge during insert, so obj is still valid */
+  }
+
   /* process subnodes */
   while (1) {
     struct hwloc__xml_import_state_s childstate;
@@ -538,14 +589,16 @@ hwloc__xml_import_object(hwloc_topology_t topology, hwloc_obj_t obj,
 
     if (!strcmp(tag, "object")) {
       hwloc_obj_t childobj = hwloc_alloc_setup_object(HWLOC_OBJ_TYPE_MAX, -1);
-      hwloc_insert_object_by_parent(topology, obj, childobj);
-      ret = hwloc__xml_import_object(topology, childobj, &childstate);
+      childobj->parent = obj; /* store the parent pointer for use in insert() below */
+      ret = hwloc__xml_import_object(topology, data, childobj, &childstate);
     } else if (!strcmp(tag, "page_type")) {
       ret = hwloc__xml_import_pagetype(topology, obj, &childstate);
     } else if (!strcmp(tag, "info")) {
       ret = hwloc__xml_import_info(topology, obj, &childstate);
     } else if (!strcmp(tag, "distances")) {
-      ret = hwloc__xml_import_distances(topology, obj, &childstate);
+      ret = hwloc__xml_import_distances(data, obj, &childstate);
+    } else if (!strcmp(tag, "userdata")) {
+      ret = hwloc__xml_import_userdata(topology, obj, &childstate);
     } else
       ret = -1;
 
@@ -558,21 +611,153 @@ hwloc__xml_import_object(hwloc_topology_t topology, hwloc_obj_t obj,
   return state->close_tag(state);
 }
 
+static int
+hwloc__xml_import_diff_one(hwloc__xml_import_state_t state,
+			   hwloc_topology_diff_t *firstdiffp,
+			   hwloc_topology_diff_t *lastdiffp)
+{
+  char *type_s = NULL;
+  char *obj_depth_s = NULL;
+  char *obj_index_s = NULL;
+  char *obj_attr_type_s = NULL;
+/* char *obj_attr_index_s = NULL; unused for now */
+  char *obj_attr_name_s = NULL;
+  char *obj_attr_oldvalue_s = NULL;
+  char *obj_attr_newvalue_s = NULL;
+
+  while (1) {
+    char *attrname, *attrvalue;
+    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+      break;
+    if (!strcmp(attrname, "type"))
+      type_s = attrvalue;
+    else if (!strcmp(attrname, "obj_depth"))
+      obj_depth_s = attrvalue;
+    else if (!strcmp(attrname, "obj_index"))
+      obj_index_s = attrvalue;
+    else if (!strcmp(attrname, "obj_attr_type"))
+      obj_attr_type_s = attrvalue;
+    else if (!strcmp(attrname, "obj_attr_index"))
+      { /* obj_attr_index_s = attrvalue; unused for now */ }
+    else if (!strcmp(attrname, "obj_attr_name"))
+      obj_attr_name_s = attrvalue;
+    else if (!strcmp(attrname, "obj_attr_oldvalue"))
+      obj_attr_oldvalue_s = attrvalue;
+    else if (!strcmp(attrname, "obj_attr_newvalue"))
+      obj_attr_newvalue_s = attrvalue;
+    else
+      return -1;
+  }
+
+  if (type_s) {
+    switch (atoi(type_s)) {
+    default:
+      break;
+    case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR: {
+      /* object attribute diff */
+      hwloc_topology_diff_obj_attr_type_t obj_attr_type;
+      hwloc_topology_diff_t diff;
+
+      /* obj_attr mandatory generic attributes */
+      if (!obj_depth_s || !obj_index_s || !obj_attr_type_s)
+	break;
+
+      /* obj_attr mandatory attributes common to all subtypes */
+      if (!obj_attr_oldvalue_s || !obj_attr_newvalue_s)
+	break;
+
+      /* mandatory attributes for obj_attr_info subtype */
+      obj_attr_type = atoi(obj_attr_type_s);
+      if (obj_attr_type == HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_INFO && !obj_attr_name_s)
+	break;
+
+      /* now we know we have everything we need */
+      diff = malloc(sizeof(*diff));
+      if (!diff)
+	return -1;
+      diff->obj_attr.type = HWLOC_TOPOLOGY_DIFF_OBJ_ATTR;
+      diff->obj_attr.obj_depth = atoi(obj_depth_s);
+      diff->obj_attr.obj_index = atoi(obj_index_s);
+      memset(&diff->obj_attr.diff, 0, sizeof(diff->obj_attr.diff));
+      diff->obj_attr.diff.generic.type = obj_attr_type;
+
+      switch (atoi(obj_attr_type_s)) {
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_SIZE:
+	diff->obj_attr.diff.uint64.oldvalue = strtoull(obj_attr_oldvalue_s, NULL, 0);
+	diff->obj_attr.diff.uint64.newvalue = strtoull(obj_attr_newvalue_s, NULL, 0);
+	break;
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_INFO:
+	diff->obj_attr.diff.string.name = strdup(obj_attr_name_s);
+	/* fallthrough */
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_NAME:
+	diff->obj_attr.diff.string.oldvalue = strdup(obj_attr_oldvalue_s);
+	diff->obj_attr.diff.string.newvalue = strdup(obj_attr_newvalue_s);
+	break;
+      }
+
+      if (*firstdiffp)
+	(*lastdiffp)->generic.next = diff;
+      else
+        *firstdiffp = diff;
+      *lastdiffp = diff;
+      diff->generic.next = NULL;
+    }
+    }
+  }
+
+  return state->close_tag(state);
+}
+
+int
+hwloc__xml_import_diff(hwloc__xml_import_state_t state,
+		       hwloc_topology_diff_t *firstdiffp)
+{
+  hwloc_topology_diff_t firstdiff = NULL, lastdiff = NULL;
+  *firstdiffp = NULL;
+
+  while (1) {
+    struct hwloc__xml_import_state_s childstate;
+    char *tag;
+    int ret;
+
+    ret = state->find_child(state, &childstate, &tag);
+    if (ret < 0)
+      return -1;
+    if (!ret)
+      break;
+
+    if (!strcmp(tag, "diff")) {
+      ret = hwloc__xml_import_diff_one(&childstate, &firstdiff, &lastdiff);
+    } else
+      ret = -1;
+
+    if (ret < 0)
+      return ret;
+
+    state->close_child(&childstate);
+  }
+
+  *firstdiffp = firstdiff;
+  return 0;
+}
+
 /***********************************
  ********* main XML import *********
  ***********************************/
 
-static void
-hwloc_xml__handle_distances(struct hwloc_topology *topology)
+static int
+hwloc_xml__handle_distances(struct hwloc_topology *topology,
+			    struct hwloc_xml_backend_data_s *data)
 {
-  struct hwloc_xml_imported_distances_s *xmldist, *next = topology->backend_params.xml.first_distances;
+  struct hwloc_xml_imported_distances_s *xmldist, *next = data->first_distances;
 
   if (!next)
-    return;
+    return 0;
 
   /* connect things now because we need levels to check/build, they'll be reconnected properly later anyway */
   hwloc_connect_children(topology->levels[0][0]);
-  hwloc_connect_levels(topology);
+  if (hwloc_connect_levels(topology) < 0)
+    return -1;
 
   while ((xmldist = next) != NULL) {
     hwloc_obj_t root = xmldist->root;
@@ -603,22 +788,28 @@ hwloc_xml__handle_distances(struct hwloc_topology *topology)
     next = xmldist->next;
     free(xmldist);
   }
+
+  return 0;
 }
 
 /* this canNOT be the first XML call */
-int
-hwloc_look_xml(struct hwloc_topology *topology)
+static int
+hwloc_look_xml(struct hwloc_backend *backend)
 {
+  struct hwloc_topology *topology = backend->topology;
+  struct hwloc_xml_backend_data_s *data = backend->private_data;
   struct hwloc__xml_import_state_s state, childstate;
   char *tag;
   hwloc_localeswitch_declare;
   int ret;
 
+  assert(!topology->levels[0][0]->cpuset);
+
   hwloc_localeswitch_init();
 
-  topology->backend_params.xml.first_distances = topology->backend_params.xml.last_distances = NULL;
+  data->first_distances = data->last_distances = NULL;
 
-  ret = topology->backend_params.xml.look(topology, &state);
+  ret = data->look_init(data, &state);
   if (ret < 0)
     goto failed;
 
@@ -626,7 +817,7 @@ hwloc_look_xml(struct hwloc_topology *topology)
   ret = state.find_child(&state, &childstate, &tag);
   if (ret < 0 || !ret || strcmp(tag, "object"))
     goto failed;
-  ret = hwloc__xml_import_object(topology, topology->levels[0][0], &childstate);
+  ret = hwloc__xml_import_object(topology, data, topology->levels[0][0], &childstate);
   if (ret < 0)
     goto failed;
   state.close_child(&childstate);
@@ -638,24 +829,114 @@ hwloc_look_xml(struct hwloc_topology *topology)
   /* we could add "BackendSource=XML" to notify that XML was used between the actual backend and here */
 
   /* if we added some distances, we must check them, and make them groupable */
-  hwloc_xml__handle_distances(topology);
-  topology->backend_params.xml.first_distances = topology->backend_params.xml.last_distances = NULL;
+  if (hwloc_xml__handle_distances(topology, data) < 0)
+    goto err;
+  data->first_distances = data->last_distances = NULL;
   topology->support.discovery->pu = 1;
 
   hwloc_localeswitch_fini();
-  return 0;
+  return 1;
 
  failed:
-  if (topology->backend_params.xml.look_failed)
-    topology->backend_params.xml.look_failed(topology);
+  if (data->look_failed)
+    data->look_failed(data);
+  if (hwloc__xml_verbose())
+    fprintf(stderr, "XML component discovery failed.\n");
+ err:
   hwloc_localeswitch_fini();
   return -1;
 }
 
+/* this can be the first XML call */
+int
+hwloc_topology_diff_load_xml(hwloc_topology_t topology __hwloc_attribute_unused,
+			     const char *xmlpath,
+			     hwloc_topology_diff_t *firstdiffp, char **refnamep)
+{
+  hwloc_localeswitch_declare;
+  char *env;
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
+
+  hwloc_localeswitch_init();
+
+  *firstdiffp = NULL;
+
+  env = getenv("HWLOC_NO_LIBXML_IMPORT");
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->import_diff(xmlpath, NULL, 0, firstdiffp, refnamep);
+  else {
+    ret = hwloc_libxml_callbacks->import_diff(xmlpath, NULL, 0, firstdiffp, refnamep);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
+  }
+
+  hwloc_localeswitch_fini();
+  return ret;
+}
+
+/* this can be the first XML call */
+int
+hwloc_topology_diff_load_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_unused,
+				   const char *xmlbuffer, int buflen,
+				   hwloc_topology_diff_t *firstdiffp, char **refnamep)
+{
+  hwloc_localeswitch_declare;
+  char *env;
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
+
+  hwloc_localeswitch_init();
+
+  *firstdiffp = NULL;
+
+  env = getenv("HWLOC_NO_LIBXML_IMPORT");
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->import_diff(NULL, xmlbuffer, buflen, firstdiffp, refnamep);
+  else {
+    ret = hwloc_libxml_callbacks->import_diff(NULL, xmlbuffer, buflen, firstdiffp, refnamep);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
+  }
+
+  hwloc_localeswitch_fini();
+  return ret;
+}
+
 /************************************************
  ********* XML export (common routines) *********
  ************************************************/
 
+#define HWLOC_XML_CHAR_VALID(c) (((c) >= 32 && (c) <= 126) || (c) == '\t' || (c) == '\n' || (c) == '\r')
+
+static int
+hwloc__xml_export_check_buffer(const char *buf, size_t length)
+{
+  unsigned i;
+  for(i=0; i<length; i++)
+    if (!HWLOC_XML_CHAR_VALID(buf[i]))
+      return -1;
+  return 0;
+}
+
 /* strdup and remove ugly chars from random string */
 static char*
 hwloc__xml_export_safestrdup(const char *old)
@@ -664,7 +945,7 @@ hwloc__xml_export_safestrdup(const char *old)
   char *dst = new;
   const char *src = old;
   while (*src) {
-    if ((*src >= 32 && *src <= 126) || *src == '\t' || *src == '\n' || *src == '\r')
+    if (HWLOC_XML_CHAR_VALID(*src))
       *(dst++) = *src;
     src++;
   }
@@ -673,93 +954,94 @@ hwloc__xml_export_safestrdup(const char *old)
 }
 
 void
-hwloc__xml_export_object (hwloc__xml_export_output_t output, hwloc_topology_t topology, hwloc_obj_t obj)
+hwloc__xml_export_object (hwloc__xml_export_state_t parentstate, hwloc_topology_t topology, hwloc_obj_t obj)
 {
+  struct hwloc__xml_export_state_s state;
   char *cpuset = NULL;
   char tmp[255];
-  unsigned nr_children = obj->memory.page_types_len + obj->infos_count + obj->distances_count + obj->arity;
   unsigned i;
 
-  output->new_child(output, "object");
-  output->new_prop(output, "type", hwloc_obj_type_string(obj->type));
+  parentstate->new_child(parentstate, &state, "object");
+
+  state.new_prop(&state, "type", hwloc_obj_type_string(obj->type));
   if (obj->os_level != -1) {
     sprintf(tmp, "%d", obj->os_level);
-    output->new_prop(output, "os_level", tmp);
+    state.new_prop(&state, "os_level", tmp);
   }
   if (obj->os_index != (unsigned) -1) {
     sprintf(tmp, "%u", obj->os_index);
-    output->new_prop(output, "os_index", tmp);
+    state.new_prop(&state, "os_index", tmp);
   }
   if (obj->cpuset) {
     hwloc_bitmap_asprintf(&cpuset, obj->cpuset);
-    output->new_prop(output, "cpuset", cpuset);
+    state.new_prop(&state, "cpuset", cpuset);
     free(cpuset);
   }
   if (obj->complete_cpuset) {
     hwloc_bitmap_asprintf(&cpuset, obj->complete_cpuset);
-    output->new_prop(output, "complete_cpuset", cpuset);
+    state.new_prop(&state, "complete_cpuset", cpuset);
     free(cpuset);
   }
   if (obj->online_cpuset) {
     hwloc_bitmap_asprintf(&cpuset, obj->online_cpuset);
-    output->new_prop(output, "online_cpuset", cpuset);
+    state.new_prop(&state, "online_cpuset", cpuset);
     free(cpuset);
   }
   if (obj->allowed_cpuset) {
     hwloc_bitmap_asprintf(&cpuset, obj->allowed_cpuset);
-    output->new_prop(output, "allowed_cpuset", cpuset);
+    state.new_prop(&state, "allowed_cpuset", cpuset);
     free(cpuset);
   }
   if (obj->nodeset && !hwloc_bitmap_isfull(obj->nodeset)) {
     hwloc_bitmap_asprintf(&cpuset, obj->nodeset);
-    output->new_prop(output, "nodeset", cpuset);
+    state.new_prop(&state, "nodeset", cpuset);
     free(cpuset);
   }
   if (obj->complete_nodeset && !hwloc_bitmap_isfull(obj->complete_nodeset)) {
     hwloc_bitmap_asprintf(&cpuset, obj->complete_nodeset);
-    output->new_prop(output, "complete_nodeset", cpuset);
+    state.new_prop(&state, "complete_nodeset", cpuset);
     free(cpuset);
   }
   if (obj->allowed_nodeset && !hwloc_bitmap_isfull(obj->allowed_nodeset)) {
     hwloc_bitmap_asprintf(&cpuset, obj->allowed_nodeset);
-    output->new_prop(output, "allowed_nodeset", cpuset);
+    state.new_prop(&state, "allowed_nodeset", cpuset);
     free(cpuset);
   }
 
   if (obj->name) {
     char *name = hwloc__xml_export_safestrdup(obj->name);
-    output->new_prop(output, "name", name);
+    state.new_prop(&state, "name", name);
     free(name);
   }
 
   switch (obj->type) {
   case HWLOC_OBJ_CACHE:
     sprintf(tmp, "%llu", (unsigned long long) obj->attr->cache.size);
-    output->new_prop(output, "cache_size", tmp);
+    state.new_prop(&state, "cache_size", tmp);
     sprintf(tmp, "%u", obj->attr->cache.depth);
-    output->new_prop(output, "depth", tmp);
+    state.new_prop(&state, "depth", tmp);
     sprintf(tmp, "%u", (unsigned) obj->attr->cache.linesize);
-    output->new_prop(output, "cache_linesize", tmp);
+    state.new_prop(&state, "cache_linesize", tmp);
     sprintf(tmp, "%d", (unsigned) obj->attr->cache.associativity);
-    output->new_prop(output, "cache_associativity", tmp);
+    state.new_prop(&state, "cache_associativity", tmp);
     sprintf(tmp, "%d", (unsigned) obj->attr->cache.type);
-    output->new_prop(output, "cache_type", tmp);
+    state.new_prop(&state, "cache_type", tmp);
     break;
   case HWLOC_OBJ_GROUP:
     sprintf(tmp, "%u", obj->attr->group.depth);
-    output->new_prop(output, "depth", tmp);
+    state.new_prop(&state, "depth", tmp);
     break;
   case HWLOC_OBJ_BRIDGE:
     sprintf(tmp, "%u-%u", obj->attr->bridge.upstream_type, obj->attr->bridge.downstream_type);
-    output->new_prop(output, "bridge_type", tmp);
+    state.new_prop(&state, "bridge_type", tmp);
     sprintf(tmp, "%u", obj->attr->bridge.depth);
-    output->new_prop(output, "depth", tmp);
+    state.new_prop(&state, "depth", tmp);
     if (obj->attr->bridge.downstream_type == HWLOC_OBJ_BRIDGE_PCI) {
       sprintf(tmp, "%04x:[%02x-%02x]",
 	      (unsigned) obj->attr->bridge.downstream.pci.domain,
 	      (unsigned) obj->attr->bridge.downstream.pci.secondary_bus,
 	      (unsigned) obj->attr->bridge.downstream.pci.subordinate_bus);
-      output->new_prop(output, "bridge_pci", tmp);
+      state.new_prop(&state, "bridge_pci", tmp);
     }
     if (obj->attr->bridge.upstream_type != HWLOC_OBJ_BRIDGE_PCI)
       break;
@@ -770,19 +1052,19 @@ hwloc__xml_export_object (hwloc__xml_export_output_t output, hwloc_topology_t to
 	    (unsigned) obj->attr->pcidev.bus,
 	    (unsigned) obj->attr->pcidev.dev,
 	    (unsigned) obj->attr->pcidev.func);
-    output->new_prop(output, "pci_busid", tmp);
+    state.new_prop(&state, "pci_busid", tmp);
     sprintf(tmp, "%04x [%04x:%04x] [%04x:%04x] %02x",
 	    (unsigned) obj->attr->pcidev.class_id,
 	    (unsigned) obj->attr->pcidev.vendor_id, (unsigned) obj->attr->pcidev.device_id,
 	    (unsigned) obj->attr->pcidev.subvendor_id, (unsigned) obj->attr->pcidev.subdevice_id,
 	    (unsigned) obj->attr->pcidev.revision);
-    output->new_prop(output, "pci_type", tmp);
+    state.new_prop(&state, "pci_type", tmp);
     sprintf(tmp, "%f", obj->attr->pcidev.linkspeed);
-    output->new_prop(output, "pci_link_speed", tmp);
+    state.new_prop(&state, "pci_link_speed", tmp);
     break;
   case HWLOC_OBJ_OS_DEVICE:
     sprintf(tmp, "%u", obj->attr->osdev.type);
-    output->new_prop(output, "osdev_type", tmp);
+    state.new_prop(&state, "osdev_type", tmp);
     break;
   default:
     break;
@@ -790,29 +1072,27 @@ hwloc__xml_export_object (hwloc__xml_export_output_t output, hwloc_topology_t to
 
   if (obj->memory.local_memory) {
     sprintf(tmp, "%llu", (unsigned long long) obj->memory.local_memory);
-    output->new_prop(output, "local_memory", tmp);
+    state.new_prop(&state, "local_memory", tmp);
   }
 
-  output->end_props(output, nr_children);
-
   for(i=0; i<obj->memory.page_types_len; i++) {
-    output->new_child(output, "page_type");
+    struct hwloc__xml_export_state_s childstate;
+    state.new_child(&state, &childstate, "page_type");
     sprintf(tmp, "%llu", (unsigned long long) obj->memory.page_types[i].size);
-    output->new_prop(output, "size", tmp);
+    childstate.new_prop(&childstate, "size", tmp);
     sprintf(tmp, "%llu", (unsigned long long) obj->memory.page_types[i].count);
-    output->new_prop(output, "count", tmp);
-    output->end_props(output, 0);
-    output->end_child(output, "page_type", 0);
+    childstate.new_prop(&childstate, "count", tmp);
+    childstate.end_object(&childstate, "page_type");
   }
 
   for(i=0; i<obj->infos_count; i++) {
     char *name = hwloc__xml_export_safestrdup(obj->infos[i].name);
     char *value = hwloc__xml_export_safestrdup(obj->infos[i].value);
-    output->new_child(output, "info");
-    output->new_prop(output, "name", name);
-    output->new_prop(output, "value", value);
-    output->end_props(output, 0);
-    output->end_child(output, "info", 0);
+    struct hwloc__xml_export_state_s childstate;
+    state.new_child(&state, &childstate, "info");
+    childstate.new_prop(&childstate, "name", name);
+    childstate.new_prop(&childstate, "value", value);
+    childstate.end_object(&childstate, "info");
     free(name);
     free(value);
   }
@@ -820,31 +1100,84 @@ hwloc__xml_export_object (hwloc__xml_export_output_t output, hwloc_topology_t to
   for(i=0; i<obj->distances_count; i++) {
     unsigned nbobjs = obj->distances[i]->nbobjs;
     unsigned j;
-    output->new_child(output, "distances");
+    struct hwloc__xml_export_state_s childstate;
+    state.new_child(&state, &childstate, "distances");
     sprintf(tmp, "%u", nbobjs);
-    output->new_prop(output, "nbobjs", tmp);
+    childstate.new_prop(&childstate, "nbobjs", tmp);
     sprintf(tmp, "%u", obj->distances[i]->relative_depth);
-    output->new_prop(output, "relative_depth", tmp);
+    childstate.new_prop(&childstate, "relative_depth", tmp);
     sprintf(tmp, "%f", obj->distances[i]->latency_base);
-    output->new_prop(output, "latency_base", tmp);
-    output->end_props(output, nbobjs*nbobjs);
+    childstate.new_prop(&childstate, "latency_base", tmp);
     for(j=0; j<nbobjs*nbobjs; j++) {
-      output->new_child(output, "latency");
+      struct hwloc__xml_export_state_s greatchildstate;
+      childstate.new_child(&childstate, &greatchildstate, "latency");
       sprintf(tmp, "%f", obj->distances[i]->latency[j]);
-      output->new_prop(output, "value", tmp);
-      output->end_props(output, 0);
-      output->end_child(output, "latency", 0);
+      greatchildstate.new_prop(&greatchildstate, "value", tmp);
+      greatchildstate.end_object(&greatchildstate, "latency");
     }
-    output->end_child(output, "distances", nbobjs*nbobjs);
+    childstate.end_object(&childstate, "distances");
   }
 
+  if (obj->userdata && topology->userdata_export_cb)
+    topology->userdata_export_cb((void*) &state, topology, obj);
+
   if (obj->arity) {
     unsigned x;
     for (x=0; x<obj->arity; x++)
-      hwloc__xml_export_object (output, topology, obj->children[x]);
+      hwloc__xml_export_object (&state, topology, obj->children[x]);
   }
 
-  output->end_child(output, "object", nr_children);
+  state.end_object(&state, "object");
+}
+
+void
+hwloc__xml_export_diff(hwloc__xml_export_state_t parentstate, hwloc_topology_diff_t diff)
+{
+  while (diff) {
+    struct hwloc__xml_export_state_s state;
+    char tmp[255];
+
+    parentstate->new_child(parentstate, &state, "diff");
+
+    sprintf(tmp, "%u", diff->generic.type);
+    state.new_prop(&state, "type", tmp);
+
+    switch (diff->generic.type) {
+    case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR:
+      sprintf(tmp, "%d", diff->obj_attr.obj_depth);
+      state.new_prop(&state, "obj_depth", tmp);
+      sprintf(tmp, "%u", diff->obj_attr.obj_index);
+      state.new_prop(&state, "obj_index", tmp);
+
+      sprintf(tmp, "%u", diff->obj_attr.diff.generic.type);
+      state.new_prop(&state, "obj_attr_type", tmp);
+
+      switch (diff->obj_attr.diff.generic.type) {
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_SIZE:
+	sprintf(tmp, "%llu", (unsigned long long) diff->obj_attr.diff.uint64.index);
+	state.new_prop(&state, "obj_attr_index", tmp);
+	sprintf(tmp, "%llu", (unsigned long long) diff->obj_attr.diff.uint64.oldvalue);
+	state.new_prop(&state, "obj_attr_oldvalue", tmp);
+	sprintf(tmp, "%llu", (unsigned long long) diff->obj_attr.diff.uint64.newvalue);
+	state.new_prop(&state, "obj_attr_newvalue", tmp);
+	break;
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_NAME:
+      case HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_INFO:
+	if (diff->obj_attr.diff.string.name)
+	  state.new_prop(&state, "obj_attr_name", diff->obj_attr.diff.string.name);
+	state.new_prop(&state, "obj_attr_oldvalue", diff->obj_attr.diff.string.oldvalue);
+	state.new_prop(&state, "obj_attr_newvalue", diff->obj_attr.diff.string.newvalue);
+	break;
+      }
+
+      break;
+    default:
+      assert(0);
+    }
+    state.end_object(&state, "diff");
+
+    diff = diff->generic.next;
+  }
 }
 
 /**********************************
@@ -855,21 +1188,28 @@ hwloc__xml_export_object (hwloc__xml_export_output_t output, hwloc_topology_t to
 int hwloc_topology_export_xml(hwloc_topology_t topology, const char *filename)
 {
   hwloc_localeswitch_declare;
-  int ret;
-#ifdef HWLOC_HAVE_LIBXML2
   char *env;
-#endif
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
 
   hwloc_localeswitch_init();
 
-#ifdef HWLOC_HAVE_LIBXML2
   env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  if (!env || !atoi(env)) {
-    ret = hwloc_libxml_export_file(topology, filename);
-  } else
-#endif
-  {
-    ret = hwloc_nolibxml_export_file(topology, filename);
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->export_file(topology, filename);
+  else {
+    ret = hwloc_libxml_callbacks->export_file(topology, filename);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
   }
 
   hwloc_localeswitch_fini();
@@ -880,21 +1220,118 @@ int hwloc_topology_export_xml(hwloc_topology_t topology, const char *filename)
 int hwloc_topology_export_xmlbuffer(hwloc_topology_t topology, char **xmlbuffer, int *buflen)
 {
   hwloc_localeswitch_declare;
-  int ret;
-#ifdef HWLOC_HAVE_LIBXML2
   char *env;
-#endif
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
 
   hwloc_localeswitch_init();
 
-#ifdef HWLOC_HAVE_LIBXML2
   env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  if (!env || !atoi(env)) {
-    ret = hwloc_libxml_export_buffer(topology, xmlbuffer, buflen);
-  } else
-#endif
-  {
-    ret = hwloc_nolibxml_export_buffer(topology, xmlbuffer, buflen);
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->export_buffer(topology, xmlbuffer, buflen);
+  else {
+    ret = hwloc_libxml_callbacks->export_buffer(topology, xmlbuffer, buflen);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
+  }
+
+  hwloc_localeswitch_fini();
+  return ret;
+}
+
+/* this can be the first XML call */
+int
+hwloc_topology_diff_export_xml(hwloc_topology_t topology __hwloc_attribute_unused,
+			       hwloc_topology_diff_t diff, const char *refname,
+			       const char *filename)
+{
+  hwloc_localeswitch_declare;
+  hwloc_topology_diff_t tmpdiff;
+  char *env;
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
+
+  tmpdiff = diff;
+  while (tmpdiff) {
+    if (tmpdiff->generic.type == HWLOC_TOPOLOGY_DIFF_TOO_COMPLEX) {
+      errno = EINVAL;
+      return -1;
+    }
+    tmpdiff = tmpdiff->generic.next;
+  }
+
+  hwloc_localeswitch_init();
+
+  env = getenv("HWLOC_NO_LIBXML_EXPORT");
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->export_diff_file(diff, refname, filename);
+  else {
+    ret = hwloc_libxml_callbacks->export_diff_file(diff, refname, filename);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
+  }
+
+  hwloc_localeswitch_fini();
+  return ret;
+}
+
+/* this can be the first XML call */
+int
+hwloc_topology_diff_export_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_unused,
+				     hwloc_topology_diff_t diff, const char *refname,
+				     char **xmlbuffer, int *buflen)
+{
+  hwloc_localeswitch_declare;
+  hwloc_topology_diff_t tmpdiff;
+  char *env;
+  int force_nolibxml;
+  int ret;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return -1;
+  }
+
+  tmpdiff = diff;
+  while (tmpdiff) {
+    if (tmpdiff->generic.type == HWLOC_TOPOLOGY_DIFF_TOO_COMPLEX) {
+      errno = EINVAL;
+      return -1;
+    }
+    tmpdiff = tmpdiff->generic.next;
+  }
+
+  hwloc_localeswitch_init();
+
+  env = getenv("HWLOC_NO_LIBXML_EXPORT");
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    ret = hwloc_nolibxml_callbacks->export_diff_buffer(diff, refname, xmlbuffer, buflen);
+  else {
+    ret = hwloc_libxml_callbacks->export_diff_buffer(diff, refname, xmlbuffer, buflen);
+    if (ret < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
   }
 
   hwloc_localeswitch_fini();
@@ -903,13 +1340,190 @@ int hwloc_topology_export_xmlbuffer(hwloc_topology_t topology, char **xmlbuffer,
 
 void hwloc_free_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_unused, char *xmlbuffer)
 {
-#ifdef HWLOC_HAVE_LIBXML2
-  char *env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  if (!env || !atoi(env)) {
-    hwloc_libxml_free_buffer(xmlbuffer);
-  } else
-#endif
-  {
-    hwloc_nolibxml_free_buffer(xmlbuffer);
+  char *env;
+  int force_nolibxml;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    return ;
   }
+
+  env = getenv("HWLOC_NO_LIBXML_EXPORT");
+  force_nolibxml = (env && atoi(env));
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    hwloc_nolibxml_callbacks->free_buffer(xmlbuffer);
+  else
+    hwloc_libxml_callbacks->free_buffer(xmlbuffer);
 }
+
+void
+hwloc_topology_set_userdata_export_callback(hwloc_topology_t topology,
+					    void (*export)(void *reserved, struct hwloc_topology *topology, struct hwloc_obj *obj))
+{
+  topology->userdata_export_cb = export;
+}
+
+static void
+hwloc__export_obj_userdata(hwloc__xml_export_state_t parentstate, int encoded,
+			   const char *name, size_t length, const void *buffer, size_t encoded_length)
+{
+  struct hwloc__xml_export_state_s state;
+  char tmp[255];
+  parentstate->new_child(parentstate, &state, "userdata");
+  if (name)
+    state.new_prop(&state, "name", name);
+  sprintf(tmp, "%lu", (unsigned long) length);
+  state.new_prop(&state, "length", tmp);
+  if (encoded)
+    state.new_prop(&state, "encoding", "base64");
+  state.add_content(&state, buffer, encoded ? encoded_length : length);
+  state.end_object(&state, "userdata");
+}
+
+int
+hwloc_export_obj_userdata(void *reserved,
+			  struct hwloc_topology *topology __hwloc_attribute_unused, struct hwloc_obj *obj __hwloc_attribute_unused,
+			  const char *name, const void *buffer, size_t length)
+{
+  hwloc__xml_export_state_t state = reserved;
+
+  if ((name && hwloc__xml_export_check_buffer(name, strlen(name)) < 0)
+      || hwloc__xml_export_check_buffer(buffer, length) < 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  hwloc__export_obj_userdata(state, 0, name, length, buffer, length);
+  return 0;
+}
+
+int
+hwloc_export_obj_userdata_base64(void *reserved,
+				 struct hwloc_topology *topology __hwloc_attribute_unused, struct hwloc_obj *obj __hwloc_attribute_unused,
+				 const char *name, const void *buffer, size_t length)
+{
+  hwloc__xml_export_state_t state = reserved;
+  size_t encoded_length;
+  char *encoded_buffer;
+  int ret;
+
+  if (name && hwloc__xml_export_check_buffer(name, strlen(name)) < 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  encoded_length = 4*((length+2)/3);
+  encoded_buffer = malloc(encoded_length+1);
+  if (!encoded_buffer) {
+    errno = ENOMEM;
+    return -1;
+  }
+
+  ret = hwloc_encode_to_base64(buffer, length, encoded_buffer, encoded_length+1);
+  assert(ret == (int) encoded_length);
+
+  hwloc__export_obj_userdata(state, 1, name, length, encoded_buffer, encoded_length);
+
+  free(encoded_buffer);
+  return 0;
+}
+
+void
+hwloc_topology_set_userdata_import_callback(hwloc_topology_t topology,
+					    void (*import)(struct hwloc_topology *topology, struct hwloc_obj *obj, const char *name, const void *buffer, size_t length))
+{
+  topology->userdata_import_cb = import;
+}
+
+/***************************************
+ ************ XML component ************
+ ***************************************/
+
+static void
+hwloc_xml_backend_disable(struct hwloc_backend *backend)
+{
+  struct hwloc_xml_backend_data_s *data = backend->private_data;
+  data->backend_exit(data);
+  free(data);
+}
+
+static struct hwloc_backend *
+hwloc_xml_component_instantiate(struct hwloc_disc_component *component,
+				const void *_data1,
+				const void *_data2,
+				const void *_data3)
+{
+  struct hwloc_xml_backend_data_s *data;
+  struct hwloc_backend *backend;
+  char *env;
+  int force_nolibxml;
+  const char * xmlpath = (const char *) _data1;
+  const char * xmlbuffer = (const char *) _data2;
+  int xmlbuflen = (int)(uintptr_t) _data3;
+  int err;
+
+  if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
+    errno = ENOSYS;
+    goto out;
+  }
+
+  if (!xmlpath && !xmlbuffer) {
+    errno = EINVAL;
+    goto out;
+  }
+
+  backend = hwloc_backend_alloc(component);
+  if (!backend)
+    goto out;
+
+  data = malloc(sizeof(*data));
+  if (!data) {
+    errno = ENOMEM;
+    goto out_with_backend;
+  }
+
+  backend->private_data = data;
+  backend->discover = hwloc_look_xml;
+  backend->disable = hwloc_xml_backend_disable;
+  backend->is_thissystem = 0;
+
+  env = getenv("HWLOC_NO_LIBXML_IMPORT");
+  force_nolibxml = (env && atoi(env));
+retry:
+  if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
+    err = hwloc_nolibxml_callbacks->backend_init(data, xmlpath, xmlbuffer, xmlbuflen);
+  else {
+    err = hwloc_libxml_callbacks->backend_init(data, xmlpath, xmlbuffer, xmlbuflen);
+    if (err < 0 && errno == ENOSYS) {
+      hwloc_libxml_callbacks = NULL;
+      goto retry;
+    }
+  }
+  if (err < 0)
+    goto out_with_data;
+
+  return backend;
+
+ out_with_data:
+  free(data);
+ out_with_backend:
+  free(backend);
+ out:
+  return NULL;
+}
+
+static struct hwloc_disc_component hwloc_xml_disc_component = {
+  HWLOC_DISC_COMPONENT_TYPE_GLOBAL,
+  "xml",
+  ~0,
+  hwloc_xml_component_instantiate,
+  30,
+  NULL
+};
+
+const struct hwloc_component hwloc_xml_component = {
+  HWLOC_COMPONENT_ABI,
+  HWLOC_COMPONENT_TYPE_DISC,
+  0,
+  &hwloc_xml_disc_component
+};
